import os
import sys
import warnings

import click


class Command(click.MultiCommand):
    def __init__(self):
        super(Command, self).__init__()

        warnings.simplefilter(action="ignore", category=FutureWarning)

    def list_commands(self, context):
        rv = []

        commands = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))

        for filename in os.listdir(commands):
            if filename.endswith(".py") and filename.startswith("command_"):
                _, name = filename.split("command_")

                name, _ = name.split(".py")

                rv.append(name)

        rv.sort()

        return rv

    def get_command(self, context, name):
        try:
            if sys.version_info[0] == 2:
                name = name.encode("ascii", "replace")

            name = "cytominer_database.commands.command_" + name

            mod = __import__(name, None, None, ["command"])
        except ImportError:
            return

        return mod.command


command = Command()



import click
import configparser
import os
import pkg_resources

import cytominer_database.ingest
import cytominer_database.munge


@click.command(
    "ingest",
    help="""\
Import CSV files into a database.

SOURCE is a directory containing subdirectories that contain CSV files.

TARGET is a connection string for the database.
"""
)
@click.argument(
    "source",
    type=click.Path(exists=True)
)
@click.argument(
    "target",
    type=click.Path(writable=True)
)
@click.option(
    "-c",
    "--config-file",
    default=pkg_resources.resource_filename(
        __name__,
        os.path.join("config", "config_htqc.ini")
    ),
    help="Configuration file.",
    type=click.Path(exists=True)
)
@click.option(
    "--munge/--no-munge",
    default=True,
    help="""\
True if the CSV files for individual compartments
have been merged into a single CSV file. If True,
the CSV will be split into one CSV per compartment.
"""
)
def command(source, target, config_file, munge):
    config = configparser.ConfigParser()

    with open(config_file, "r") as config_fd:
        config.read_file(config_fd)

    if munge:
        cytominer_database.munge.munge(config=config, source=source)

    cytominer_database.ingest.seed(source, target, config)



import setuptools


setuptools.setup(
    name='cytominer_database',
    version='0.0.1b3',
    author="Shantanu Singh",
    author_email="shsingh@broadinstitute.org",
    entry_points="""
    [console_scripts]
    cytominer-database=cytominer_database.command:command
    """,
    long_description="cytominer-database provides mechanisms to import CSV "
                     "files generated in a morphological profiling experiment "
                     "into a database backend. "
                     "Please refer to the online documentation at "
                     "http://cytominer-database.readthedocs.io",
    package_data={
        'cytominer_database': [
            'config/*.ini',
            'config/*.json',
            'config/*.sql'
        ],
    },
    packages=setuptools.find_packages(
        exclude=[
            'tests',
            'doc'
        ]
    ),
    include_package_data=True,
    install_requires=[
        'backports.tempfile>=1.0rc1',
        'click>=6.7',
        'configparser>=3.5.0',
        'csvkit>=1.0.2',
        'networkx<2.0',  # TODO: https://github.com/blaze/odo/issues/579 -- remove dependency when resolved
        'odo>=0.5.0',
        'pandas>=0.20.3'
    ],
    license='BSD',
    url='https://github.com/cytomining/cytominer-database'
)



"""

"""

import click
import cytominer_database.utils
import os
import pandas as pd


def munge(config, source, target=None):
    """ Searches ``source`` for directories containing a CSV file corresponding to
    per-object measurements, then splits the CSV file into one CSV file per compartment.
    For instance, the CSV file may comprise of measurements combined across Cells,
    Cytoplasm, and Nuclei. ``munge`` will split this CSV file into 3 CSV files:
    Cells.csv, Cytoplasm.csv, and Nuclei.csv.

    :param config: Configuration file.

    :param source: Directory containing subdirectories that contain an object CSV file.

    :param target: Output directory. If not specified, then it is same as ``source``.

    :return: list of subdirectories that have an object CSV file.

    """

    if not target:
        target = source

    directories = sorted(list(cytominer_database.utils.find_directories(source)))

    valid_directories = [] # list of subdirectories that have an object CSV file.

    for directory in directories:
        try:
            obj = pd.read_csv(os.path.join(directory, config["filenames"]["object"]), header=[0, 1])

        except IOError as e:
            click.echo(e)

            continue

        valid_directories.append(directory)

        target_directory = directory.replace(source, target)

        if not os.path.exists(target_directory):
            os.makedirs(target_directory)

        for compartment_name in set(obj.columns.get_level_values(0).tolist()) - {'Image'}:

            # select columns of the compartment
            compartment = pd.concat([obj['Image'], obj[compartment_name]], axis=1)

            # Create a new column
            compartment['ObjectNumber'] = compartment['Number_Object_Number']

            cols = compartment.columns.tolist()

            # Move ImageNumber and ObjectNumber to the front

            cols.insert(0, cols.pop(cols.index('ObjectNumber')))

            cols.insert(0, cols.pop(cols.index('ImageNumber')))

            compartment = compartment.loc[:, cols]

            compartment.to_csv(os.path.join(target_directory, compartment_name + '.csv'), index=False)

    return valid_directories



import csv
import glob
import logging
import os
import tempfile

import csvkit.utilities.csvclean


logger = logging.getLogger(__name__)


def find_directories(directory):
    """List subdirectories.

    :param directory: directory

    :return: list of subdirectories of ``directory``

    """
    directories = set()

    filenames = glob.glob(os.path.join(directory, '*/'))

    for filename in filenames:
        pathname = os.path.relpath(filename)

        directories.add(pathname)

    return directories


def validate_csv(csvfile):
    """Validate CSV file.

    The CSV file typically corresponds to either a measurement made on a compartment, e.g. Cells.csv, or on an image,
    e.g. Image.csv. The validation performed is generic - it simply checks for malformed CSV files.

    This uses csvclean to check for validity of a CSV file.

    :param csvfile: CSV file to validate

    :return: True if valid, False otherwise.

    """
    if os.stat(csvfile).st_size == 0:
        return False

    with open(csvfile, "r") as csvfd:
        reader = csv.reader(csvfd)

        nrows = sum(1 for _ in reader) - 1  # exclude header

    if nrows < 1:
        return False

    with tempfile.TemporaryFile(mode="w+") as tmpfile:
        csvclean = csvkit.utilities.csvclean.CSVClean(args=["-n", csvfile], output_file=tmpfile)

        csvclean.run()

        tmpfile.seek(0)

        csvcheck = tmpfile.read().strip()

    return csvcheck == "No errors."


def validate_csv_set(config, directory):
    """Validate a set of CSV files.

    This function validates a set of CSV files in a directory. These CSV files correspond to measurements
    made on different cellular compartments, e.g. Cells.csv, Cytoplasm.csv, Nuclei.csv. An Image.csv file,
    corresponding to measurements made on the whole image, along with metadata, is also typically present.

    :param config: configuration file - this contains the set of CSV files to validate.
    :param directory: directory containing the CSV files.

    :return: a tuple where the first element is the list of pattern CSV files, the second is the image CSV file.

    """
    image_csv = os.path.join(directory, config["filenames"]["image"])

    if not os.path.isfile(image_csv):
        raise IOError("{} not found in {}. Skipping.".format(config["filenames"]["image"], directory))

    pattern_csvs = [filename for filename in glob.glob(os.path.join(directory, '*.csv')) if filename not in [
        os.path.join(directory, config['filenames']['image']),
        os.path.join(directory, config['filenames']['object']),
        os.path.join(directory, config['filenames']['experiment'])
    ]]

    filenames = pattern_csvs + [image_csv]

    file_checks = dict({(filename, validate_csv(filename)) for filename in filenames})

    if not all(file_checks.values()):
        invalid_files = ",".join([os.path.basename(filename) for (filename, valid) in file_checks.items() if not valid])
        raise IOError("Some files were invalid: {}. Skipping {}.".format(invalid_files, directory))

    return pattern_csvs, image_csv
